#lang scheme

(define table
  `([main [("apple") ,(lambda () 'apple)]
          [("banana") ,(lambda () 'banana)]
          [(fruit-salad "!") ,(lambda (x)
                                `(,x !))]]
    [fruit-salad [(fruit fruit) ,(lambda (x y)
                                   (list x y))]
                 [(fruit fruit fruit) ,(lambda (x y z)
                                         (list x y z))]]
    [fruit [("orange") ,(lambda () 'orange)]
           [("grape") ,(lambda () 'grape)]]))

;; ----------------------------------------
;; Combinators:

(define ((or-k try next) success-k fail-k)
  (local [(define (make-fst-success-k success-k fail-k)
            (lambda (val stream next-fst)
              (success-k val stream
                         (lambda (success-k fail-k)
                           (next-fst (make-fst-success-k success-k fail-k)
                                     (make-fst-fail-k success-k fail-k))))))
          (define (make-fst-fail-k success-k fail-k)
            (lambda ()
              (next success-k fail-k)))]
    (try (make-fst-success-k success-k fail-k)
         (make-fst-fail-k success-k fail-k))))

(define ((and-k first second combine) success-k fail-k)
  (first (local [(define (make-fst-success-k success-k fail-k)
                   (lambda (fst-vals stream fst-next)
                     ((second stream)
                      (make-snd-success-k fst-vals fst-next success-k fail-k)
                      (lambda ()
                        (fst-next (make-fst-success-k success-k fail-k)
                                  fail-k)))))
                 (define (make-snd-success-k fst-vals fst-next success-k fail-k)
                   (lambda (snd-vals stream snd-next)
                     (success-k 
                      (combine fst-vals snd-vals)
                      stream
                      (lambda (success-k fail-k)
                        (snd-next 
                         (make-snd-success-k fst-vals fst-next success-k fail-k)
                         (lambda ()
                           (fst-next (make-fst-success-k success-k fail-k)
                                     fail-k)))))))]
           (make-fst-success-k success-k fail-k))
         fail-k))

(define (wrap-success-k try val->val)
  (local [(define (make-my-success-k success-k fail-k)
            (lambda (val stream next)
              (success-k (val->val val) stream
                         (lambda (success-k fail-k)
                           (next (make-my-success-k success-k fail-k)
                                 fail-k)))))]
    (lambda (success-k fail-k)
      (try (make-my-success-k success-k fail-k) fail-k))))

;; ----------------------------------------

(define (parse what table stream)
  (local [(define defn (assq what table))]
    (if defn
        (parse-alts (rest defn) table stream)
        (error 'parse "no definition for ~a" what))))

(define (parse-alts alts table stream)
  (if (null? alts)
      (lambda (success-k fail-k) (fail-k))
      (or-k (parse-case (first alts) table stream)
            (parse-alts (rest alts) table stream))))
        
(define (parse-case case table stream)
  (local [(define seq (first case))
          (define handler (second case))]
    (wrap-success-k (parse-sequence seq table stream)
                    (lambda (vals) (apply handler vals)))))

(define (parse-sequence seq table stream)
  (cond
    [(empty? seq) 
     (lambda (success-k fail-k)
       (success-k empty stream
                  (lambda (success-k fail-k)
                    (fail-k))))]
    [else 
     (and-k (parse-item (first seq) table stream)
            (lambda (stream) (parse-sequence (rest seq) table stream))
            append)]))

(define (parse-item item table stream)
  (cond
    [(string? item)
     (if (and (cons? stream)
              (equal? item (first stream)))
         (lambda (success-k fail-k)
           (success-k empty (rest stream) 
                      (lambda (success-k fail-k)
                        (fail-k))))
         (lambda (success-k fail-k)
           (fail-k)))]
    [else
     (wrap-success-k (parse item table stream)
                     list)]))

((parse 'fruit-salad table '("orange" "orange" "!"))
 (lambda (val stream again) val)
 (lambda () 'nope!))

((parse 'main table '("orange" "orange" "orange" "!"))
 (lambda (val stream again) val)
 (lambda () 'nope!))
